/* 
   (c) Copyright 2001-2002  convergence integrated media GmbH.
   (c) Copyright 2002-2005  convergence GmbH.

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lite_internal.h"

#include "font.h"
#include "label.h"

D_DEBUG_DOMAIN( Lite_Label, "LiTE/Label", "LiteLabel");

struct _LiteLabel {
     LiteBox             box;

     LiteFont           *font;
     char               *text;
     DFBColor            text_color;

     LiteLabelAlignment  alignment;
};

static void draw_label (LiteBox *box, const DFBRegion *region, DFBBoolean clear);
static void destroy_label (LiteBox *box);

LiteLabel *lite_new_label (LiteBox *parent,
                           int x, int y, int width, int font_size)
{
     LiteFont  *font;
     LiteLabel *label;
     int        height;
     DFBResult  result;
     IDirectFBFont *font_interface;

     result = lite_get_font ("default", LITE_FONT_PLAIN, font_size, DEFAULT_FONT_ATTRIBUTE, &font);
     if (result != DFB_OK)
          return NULL;

     result = lite_font(font, &font_interface);
     if (result != DFB_OK)
          return NULL;
          
     font_interface->GetHeight (font_interface, &height);

     label = calloc (1, sizeof(LiteLabel));

     label->box.parent = parent;

     label->box.rect.x = x;
     label->box.rect.y = y;
     label->box.rect.w = width;
     label->box.rect.h = height;

     /* default black font color */
     label->text_color.r = 0x00;
     label->text_color.g = 0x00;
     label->text_color.b = 0x00;
     label->text_color.a = 0xff;

     if (lite_init_box (LITE_BOX (label))) {
          lite_release_font (font);
          free (label);
          return NULL;
     }

     label->box.Draw    = draw_label;
     label->box.Destroy = destroy_label;

     label->font = font;
     label->text = strdup ("");

     return label;
}

DFBResult lite_set_label_text (LiteLabel  *label,
                               const char *text)
{
     if (!strcmp (label->text, text))
          return DFB_OK;

     free (label->text);

     label->text = strdup (text);

     return lite_update_box (LITE_BOX (label), NULL);
}

DFBResult lite_set_label_alignment (LiteLabel          *label,
                                    LiteLabelAlignment  alignment)
{
     if (label->alignment == alignment)
          return DFB_OK;

     label->alignment = alignment;

     return lite_update_box (LITE_BOX (label), NULL);
}

DFBResult lite_set_label_font (LiteLabel     *label,
                               const char    *spec,
                               LiteFontStyle  style,
                               int            size,
                               DFBFontAttributes attr)
{
     LiteFont *font;
     DFBResult result;
     IDirectFBFont *font_interface;

     result = lite_get_font( spec, style, size, attr, &font );
     if (result != DFB_OK)
          return result;

     result = lite_font(font, &font_interface);
     if (result != DFB_OK)
          return result;

     lite_release_font( label->font );

     label->font = font;

     font_interface->GetHeight (font_interface, &label->box.rect.h);

     return lite_update_box (LITE_BOX (label), NULL);
}

/* file internals */

static void destroy_label (LiteBox *box)
{
     LiteLabel *label = LITE_LABEL (box);

     free (label->text);

     lite_release_font (label->font);

     lite_destroy_box (box);
}

static void draw_label (LiteBox *box, const DFBRegion *region, DFBBoolean clear)
{
     int                  x       = 0;
     DFBSurfaceTextFlags  flags   = DSTF_TOP;
     IDirectFBSurface    *surface = box->surface;
     LiteLabel           *label   = LITE_LABEL (box);
     IDirectFBFont       *font_interface;
     DFBResult            result;

     result = lite_font(label->font, &font_interface);
     if (result != DFB_OK)
          return;

     surface->SetFont (surface, font_interface);
     surface->SetClip (surface, region);

     /* Fill the background */
     if (clear)
          lite_clear_box (box, region);

     switch (label->alignment) {
          case LITE_LABEL_LEFT:
               flags |= DSTF_LEFT;
               break;

          case LITE_LABEL_RIGHT:
               x = box->rect.w - 1;
               flags |= DSTF_RIGHT;
               break;

          case LITE_LABEL_CENTER:
               x = box->rect.w / 2;
               flags |= DSTF_CENTER;
               break;
     }

     /* Draw the text */
     surface->SetColor (surface, 
                         label->text_color.r, label->text_color.g,
                         label->text_color.b, label->text_color.a);
     surface->DrawString (surface, label->text, -1, x, 0, flags);
}

DFBResult lite_set_label_color (LiteLabel *label, DFBColor *color)
{
     DFBResult res; 

     if(!color && !label)
          return DFB_INVARG;
     
     if (!DFB_COLOR_EQUAL(label->text_color, *color)) {
               label->text_color = *color; 
               res = lite_update_box( LITE_BOX(label), NULL);
     }
     else
          res = DFB_OK;

     return res;
}
